/**
* This script provides compatibility functions so we (GraalJS, but also Nashorn, ...) can execute the V8 test suite.
* Note that the functions have been renamed from e.g. %DeoptimizeFunction to v8DeoptimizeFunction, as "%" is no valid part of a function name.
*/

//this method is used e.g. to replace checks like assertTrue(v8HasFastProperties()) that we cannot check properly
function v8AssertIGNORE(a) {
	return undefined;
}
	 
function v8OptimizeFunctionOnNextCall() {
	return undefined;
}

function v8DeoptimizeFunction() {
	return undefined;
}

function v8NeverOptimizeFunction() {
	return undefined;
}

function v8ClearFunctionTypeFeedback() {
	return undefined;
}

function v8RunMicrotasks() {
	return undefined;
}

function v8EnqueueMicrotask(a) {
	return undefined;
}

function v8DebugPrint() {
	for (var i in arguments) {
		print(arguments[i]);
	}
	return undefined;
}

function v8DebugPrintScopes() {
	return undefined;
}

function v8ArgumentsLength(arg) {
	return arg.length;
}

function v8Arguments(arg,i) {
	return arg[i];
}

function v8NotifyContextDisposed() {
	return true;
}

function v8IsConcurrentRecompilationSupported() {
	return true;
}

function v8GetOptimizationStatus() {
	return 0;
}

function v8ToFastProperties(obj) {
	return obj;
}

function v8FlattenString(str) {
	return str;
}

function v8IsValidSmi(value) {
	return Debug.class(value) === "java.lang.Integer";
}

function v8HasFastObjectElements() {
	return true;
}

function v8HasFastSmiOrObjectElements() {
	return true;
}

function v8HasFastSmiElements(array) {
	return Debug.arraytype(array).contains("Int") || Debug.arraytype(array).contains("Byte") || Debug.arraytype(array).contains("Empty") || v8HasFastSmiElementsSlow(array);
}

//helper function
function v8HasFastSmiElementsSlow(array) {
	for (var i=0;i<array.length;i++) {
		if (array.hasOwnProperty(i)) {
			var type = Debug.class(array[i]);
			if (!(type === "java.lang.Integer" || type === "java.lang.Byte")) {
				return false;
			}
		}
	}
	return true;
}

function v8HasFastDoubleElements(array) {
	return Debug.arraytype(array).contains("Double") || v8HasFastSmiElements(array) || v8HasFastDoubleElementsSlow(array);
}

//helper function
function v8HasFastDoubleElementsSlow(array) {
	for (var i=0;i<array.length;i++) {
		if (array.hasOwnProperty(i)) {
			var type = Debug.class(array[i]);
			if (!(type === "java.lang.Double" || type === "java.lang.Integer" || type === "java.lang.Byte")) {
				return false;
			}
		}
	}
	return true;
}

function v8HasFastHoleyElements(obj) {
	return Debug.isHolesArray(obj);
}

function v8HasFastProperties() {
	return true;
}

function v8HasDictionaryElements() {
	return true;
}

function v8HaveSameMap(obj1, obj2) {
	return Debug.haveSameMap(obj1, obj2);
}

function v8HasExternalArrayElements() {
	return true;
}

function v8HasExternalByteElements() {
	return true;
}

function v8HasExternalUnsignedByteElements() {
	return true;
}

function v8HasFixedUint8Elements(ob) {
	return true;
}

function v8HasFixedInt16Elements(ob) {
	return true;
}

function v8HasFixedUint16Elements(ob) {
	return true;
}

function v8HasFixedInt32Elements(ob) {
	return true;
}

function v8HasFixedUint32Elements(ob) {
	return true;
}

function v8HasFixedFloat32Elements(ob) {
	return true;
}

function v8HasFixedFloat64Elements(ob) {
	return true;
}

function v8HasFixedUint8ClampedElements(ob) {
	return true;
}

function gc() {
	return undefined;
}

function v8IsMinusZero(a) {
	if (a === 0 && typeof(a) === "number") {
		return (1/a) === -Infinity;
	} else {
		return false;
	}
}

function v8RunningInSimulator() {
	return false;
}

function v8StringCharCode(st, idx) {
	return st.charAt(idx);
}

function v8StringCharCodeAt(st, idx) {
	return st.charCodeAt(idx);
}

function v8StringFromCharCode(code) {
	return String.fromCharCode(code);
}

function v8UnblockConcurrentRecompilation() {
	return undefined;
}

function v8IsObject(obj) {
	return typeof(obj) === "object" || typeof(obj) === "undefined";
}

function v8GetOptimizationCount(obj) {
	return 0;
}

function v8MaxSmi() {
	return 2147483647;
}

function v8DebugDisassembleFunction() {
	return undefined;
}

function v8SetProperty(obj, name, value) {
	obj[name] = value;
}

function v8SetAllocationTimeout() {
	return undefined;
}

function v8OptimizeObjectForAddingMultipleProperties(obj) {
	return obj;
}

function v8TryMigrateInstance(obj) {
	return obj;
}

function v8StringParseInt(str, rad) {
	return parseInt(str,rad);
}

function externalizeString(str) {
	return str;
}

function v8CallFunction() {
	var thisObj = arguments[0];
	var func = arguments[arguments.length-1];
	var newArgs = [];
	for (var i=1;i<arguments.length-1;i++) {
		newArgs[i-1] = arguments[i];
	}
	return func.call(thisObj,...newArgs);
}

function v8Call() {
	var func = arguments[0];
	var thisObj = arguments[1];
	var newArgs = [];
	for (var i=2;i<arguments.length;i++) {
		newArgs[i-2] = arguments[i];
	}
	return func.call(thisObj,...newArgs);
}

function v8Apply() {
	var thisObj = arguments[0];
	var func = arguments[1];
	var newArgs = [];
	for (var i=2;i<arguments.length;i++) {
		newArgs[i-2] = arguments[i];
	}
	return func.apply(thisObj,...newArgs);
}


function v8GlobalParseInt(value, radix) {
	return parseInt(value, radix);
}

function v8SetFlags(flags) {
	return undefined;
}

function v8Break() {
	return undefined;
}

function v8DebugBreak() {
	return undefined;
}

function v8NewString(len, flag) {
	return new String(len);
}

function v8ValueOf(obj) {
	return obj.valueOf();
}

function v8ClassOf(obj) {
	return Debug.className(obj);
}

function v8ObjectFreeze(obj) {
	return obj.freeze();
}

function v8StringCharFromCode(str) {
	return str;
}

function v8SmiLexicographicCompare(a, b) {
	return Debug.stringCompare(a+"",b+"");
}

function v8StringCompare(a, b) {
	return Debug.stringCompare(a+"",b+"");
}

function v8StringCompareRT(a, b) {
	return v8StringCompare(a, b);
}

function v8HasFixedInt8Elements() {
	return false;
}

function v8MakeReferenceError(message) {
	return new ReferenceError(message);
}

function v8AbortJS(message) {
	exit(message);
}

function v8HomeObjectSymbol() {
	return Symbol("__home__");
}

function v8PreventExtensions(obj) {
	return Object.preventExtensions(obj);
}

function v8GetPropertyNames(obj) {
	return Object.keys(obj);
}

function v8NormalizeElements(arr) {
	return arr;
}

function v8SymbolIsPrivate(sym) {
	return false;
}

function v8CreatePrivateSymbol(sym) {
	return Symbol(sym);
}

function v8CreatePrivateOwnSymbol(sym) {
	return Symbol(sym);
}

function v8ArrayBufferNeuter(arr) {
	return arr;
}

function v8GetScript(name) {
	return undefined;
}

function v8AddNamedProperty(object, name, valueParam, prop) {
	var isWritable = !(prop & 1);
	var isEnumerable = !(prop & 2);
	var isConfigurable = !(prop & 4); 
	
	var propDesc = { value: valueParam, writable: isWritable, enumerable: isEnumerable, configurable: isConfigurable };
	
	return Object.defineProperty(object, name, propDesc);
}

function v8SetValueOf(a,b) {
	return a;
}

function v8OneByteSeqStringSetChar() {
}

function v8TwoByteSeqStringSetChar() {
}

function v8IsRegExp(obj) {
	return Object.prototype.toString.call(obj) == "[object RegExp]";
}

function v8IsArray(obj) {
	return Object.prototype.toString.call(obj) == "[object Array]";
}

function v8IsFunction(obj) {
	return Object.prototype.toString.call(obj) == "[object Function]";
}

function v8IsSpecObject(obj) {
	return Object.prototype.toString.call(obj) == "[object Date]";
}

function v8FunctionGetScript(scr) {
}

function v8ConstructDouble(hi,lo) {
	return Debug.constructDouble(hi,lo);
}

function v8DoubleHi(value) {
	return Debug.doubleHi(value);
}

function v8DoubleLo(value) {
	return Debug.doubleLo(value);
}

function v8HasSloppyArgumentsElements(a) {
	return true;
}

function v8MakeError(i,msg) {
	return new Error(msg);
}

function v8AllocateHeapNumber() {
	return 0;
}

function v8DeoptimizeNow() {
	Debug.deoptimize();
}

function v8OptimizeOsr() {
	return undefined;
}

function v8DisassembleFunction(f) {
	Debug.deoptimize(); //not what is expected	
	return undefined;
}

function v8IsStrong(a) {
	return Debug.isStrong(a);
}

var UNDETECTABLE = {};
function v8GetUndetectable() {
	return UNDETECTABLE;
}

function v8FixedArrayGet(vector, slot) {
	return vector[slot];
}

function v8GetTypeFeedbackVector(vector) {
	return undefined;
}

function v8DebugGetLoadedScripts() {
	return undefined;
}

function v8IsSmi(value){
	return typeof(value)==="number" && Math.floor(value) === value && !v8IsMinusZero(value) && Debug.class(value) === "java.lang.Integer" && value != 2147483648;
}

function v8FunctionGetInferredName(a) {
	return a.name;
}

function v8TruncateString(string, index) {
	return string.substring(0,index);
}

function v8MathClz32(a) {
	return Math.clz32(a);
}

function v8ScheduleBreak() {
	return undefined;
}

function v8FormatMessageString(i, m, o, p) {
	if (i < 0) { throw new TypeError("out of bounds"); }
	return new Error(m+o+p);
}

function v8ToMethod(f, o) {
	return f.toMethod(o);
}

function v8RegExpConstructResult() {
	return undefined;
}

function v8DefineAccessorPropertyUnchecked() {
	throw new Error("illegal access");
}

function v8DefineDataPropertyUnchecked() {
	throw new Error("illegal access");
}

function v8NewObjectFromBound(obj) {
	return undefined;
}

function v8IsConstructCall(obj) {
	return false;
}

function v8DeoptimizeNow() {
	return undefined;
}

function v8AtomicsFutexNumWaitersForTesting(a, b) {
	return 0;
}

function v8ToLength(a) {
	return Debug.toLength(a);
}

function v8ToName(a) {
	return Debug.toName(a);
}

function v8ToString(a) {
	return Debug.toStringConv(a);
}

function v8ToPrimitive(a) {
	return Debug.toPrimitive(a);
}

function v8ToPrimitive_Number(a) {
	return Debug.toPrimitiveNumber(a);
}

function v8ToPrimitive_String(a) {
	return Debug.toPrimitiveString(a);
}

function v8ToNumber(a) {
	return Debug.toNumber(a);
}

function v8GetHoleNaNUpper() {
	return 0;
}

function v8GetHoleNaNLower() {
	return 2146959360;
}

function v8TailCall() {
	return undefined;
}

function v8IsJSReceiver(a) {
	return typeof(a) === "object";
}

function v8FunctionGetSourceCode() {
}

function v8SetForceInlineFlag() {
}

function v8MathSqrt(a) {
        return Math.sqrt(a);
}

function v8IsAsmWasmCode() {
	return true;
}

function v8IsNotAsmWasmCode() {
	return true;
}

function v8ExecuteInDebugContext() {
	return undefined;
}

function v8SpeciesProtector() {
	return false;
}

function v8GeneratorGetFunction() {
	return undefined;
}

function v8BaselineFunctionOnNextCall() {
	return undefined;
}

function v8InterpretFunctionOnNextCall() {
	return undefined;
}

function v8CreateDataProperty(obj, key, val) {
	obj[key] = val;
}

